this.name		 = "telescope_debug";
this.author		 = "cag";
this.copyright	 = "2017 cag";
this.licence	 = "CC BY-NC-SA 3.0";
this.description = "debugging helper fns";
this.version	 = "1.0";

/* jshint elision:true, shadow:false, esnext: true, curly: false, maxerr: 1000, asi: true, laxbreak: true, undef: true, unused:true
*/
/* global log, worldScripts, Script, Vector3D, Quaternion
*/

(function(){
/* validthis: true
*/
"use strict";

// false prevents output, though functions are still tracked
// - except with start's header, if it's forced by a child fn's msg or by its stop's msg.
// So putting text in a stop will force output of start's header when there's no regular
// output from add, begin or end
// - useful for exceptional situations, where a fn quits early: put a stop w/ msg before the return
//	 and output start's header as well; only see the fn output when something interesting happens
this.$fns_watched = {
'shipSpawned':				false,
'shipScoopedOther':			false,
'update_one_Sighting':		true,
'_update_some_Sighting':	true,
'_update_Sightings':		false,
'_delete_Sighting':			true,
'_add_Sighting':			true
};

// worldScripts.telescope.$cag.init();

this.$project_name = 'telescope';

this.$DEBUG_CONSOLE_WIDTH = 200;
this.$DEBUG_OFF = true;
//this.$DEBUG_OFF = false;

this._display_map = function _display_map( map ) {
	var that = _display_map;
	var _showProps = ( that._showProps = that._showProps || worldScripts.telescope_debug._showProps );
	
	if( map.hasOwnProperty( 'map' ) ) {
		map = map.map;
	}
	var tmp = _showProps( map, 'map', false, true, true );
	const MFD_FRIENDLY = 1,		// bounty === 0 && !markedForFines
		  MFD_UNSOCIABLE = 2,	// bounty || markedForFines
		  MFD_ACTIVE = 4,		// has .target || defenseTargets.length > 0
		  MFD_HOSTILE = 8,		// in_ents_Targets || targeting_ps
		  	MFD_ATTITUDE = 15,	// those of 1st 4 flags used to choose targets
		  MFD_NEARBY = 16,		// distance < scannerRange
		  MFD_PROTECTED = 32,	// .withinStationAegis
		  MFD_FARAWAY = 64,		// distance > scannerRange
		  	MFD_RANGED = 112;	// those of prev. 3 flags used to limit those chosen
	const MFD_SALVAGE = 1,		// cargo, escape pods, derelicts
		  MFD_MINING = 2,		// asteroids, boulders, splinters & metal fragments
		  MFD_WEAPONS = 4,		// mines & missiles
//		  	MFD_INANIMATE = 7,	// those of 1st 3 flags excluded from dynamic filtering
		  MFD_TRADERS = 8,		// ships .isTrader & escorts
		  MFD_POLICE = 16,		// scanClass === 'CLASS_POLICE'
		  MFD_PIRATES = 32,		// .isPirate & .isPirateVictim
		  MFD_MILITARY = 64,	// scanClass === 'CLASS_MILITARY'
		  MFD_ALIENS = 128,		// scanClass === 'CLASS_THARGOID'
		  MFD_NEUTRAL = 256,	// scanClass === 'CLASS_NEUTRAL' and not in any above category (e.g., miners, hunters, etc.)
//		  	MFD_ALLSHIPS = 504,	// all of the previous 6
		  MFD_STATION = 512,	// .isStation
		  MFD_NAVIGATION = 1024,// some stations & beacons (may include a ship if emitting a beacon)
		  MFD_CELESTIAL = 2048;	// sun, planets, moons
//		  	MFD_ORIENT = 3584;	// all of the previous 3

	tmp += '\ndynamic flags';
	var flags = map.dynamicMFD;
	if( flags ) {
		if( flags & MFD_FRIENDLY ) tmp += ':  FRIENDLY';
		if( flags & MFD_UNSOCIABLE ) tmp += ':	UNSOCIABLE';
		if( flags & MFD_ACTIVE ) tmp += ':	ACTIVE';
		if( flags & MFD_HOSTILE ) tmp += ':	 HOSTILE';
		if( flags & MFD_NEARBY ) tmp += ':	NEARBY';
		if( flags & MFD_FARAWAY ) tmp += ':	 FARAWAY';
		if( flags & MFD_PROTECTED ) tmp += ':  PROTECTED';
	}
	tmp += '\nstatic flags';
	flags = map.staticMFD;
	if( flags ) {
		if( flags & MFD_SALVAGE ) tmp += ':	 SALVAGE';
		if( flags & MFD_MINING ) tmp += ':	MINING';
		if( flags & MFD_WEAPONS ) tmp += ':	 WEAPONS';
		if( flags & MFD_TRADERS ) tmp += ':	 TRADERS';
		if( flags & MFD_POLICE ) tmp += ':	POLICE';
		if( flags & MFD_PIRATES ) tmp += ':	 PIRATES';
		if( flags & MFD_MILITARY ) tmp += ':  MILITARY';
		if( flags & MFD_ALIENS ) tmp += ':	ALIENS';
		if( flags & MFD_STATION ) tmp += ':	 STATION';
		if( flags & MFD_NAVIGATION ) tmp += ':	NAVIGATION';
		if( flags & MFD_CELESTIAL ) tmp += ':  CELESTIAL';
	}
	log( tmp );
}

this._dump_map = function _dump_map() {		// dump Sightings list for telescope
	var that = _dump_map;
	var cd = ( that.cd = that.cd || worldScripts.telescope_debug );
	var ws = ( that.ws = that.ws || worldScripts.telescope );
	var curr_S = ( that.curr_S = that.curr_S || ws.$curr_Sighting );
	
	var pst = curr_S.ent || null;
	var mapping = ws.$SightingsMap, maplen = mapping.length;
	var map, ent, target, thargoids = '';
	var out = '____ent_dist.rank   index	 rel. dir.n		 person	   [entity description minus position, scanClass & status labels ...]	 [ve_colour,  lb_size,	ml_size]\n';
	// var out = '____ent_dist.rank   index	 rel. dir.n	   curr.grav.D	 person	   [entity description minus position, scanClass & status labels ...]	 [ve_colour,  lb_size,	ml_size]\n';
	var i, padding = '________________';
	var	 ent_dist, rank, index, rel_dir, gs_curr, entID, descrn, gs_maxed;
	for( i = 0; i < maplen; i++ ) {
		map = mapping[i]; ent = map.ent;
		gs_curr = map.gs_curr_dist;
		gs_maxed = gs_curr > 0 && gs_curr === map.gs_max_dist;
		target = pst && pst === ent;
		if(ent.isThargoid) { out += ent+'\n';
			thargoids += '.	   dataKey = '+map.dataKey+', last_posn = '+map.last_posn+' roles = '+map.roles+'\n';
		}
		ent_dist = map.ent_dist;
		if( ent_dist > 999999999 )
			ent_dist = cd._number_str( ent_dist / 1000000000 ) + ' e9';
		else
			ent_dist = cd._number_str( ent_dist );
		ent_dist = padding.slice( ent_dist.length-10 )+ ent_dist;
		if( target ) {
			// ent_dist = ent_dist.replace(/^_*/, '==>');
			ent_dist = ent_dist[1] == '_' ? '=>' + ent_dist.substring(2) :
					   ent_dist[0] == '_' ? '|>' + ent_dist.substring(1) :
					   '>' + ent_dist;
		}
		rank = map.rank.toUpperCase();
		index = i < 10 ? '__' + i : i > 99 ? i : '_' + i;
		rel_dir = ws._relativeDirection( ent.position, map );		
		rel_dir = rel_dir ? padding.slice( rel_dir.length-9 ) + rel_dir : '???';
		// gs_curr = cd._number_str( gs_curr );							
		// gs_curr = padding.slice( gs_curr.length-(gs_maxed ? 10 : 11) ) + gs_curr;
		entID = ent.entityPersonality;
		entID = entID === undefined ? 'xxxxx' : entID.toString();	
		entID = padding.slice( entID.length-6 ) + entID;
		descrn = ent.toString();
		descrn = descrn.replace(/position\: /, '' );
		descrn = descrn.replace(/scanClass\: /, '__' );
		descrn = descrn.replace(/CLASS_/, '' );
		descrn = descrn.replace(/status\: /, '__' );
		descrn = descrn.replace(/STATUS_/, '' );
		out += ent_dist+'_'+rank
		out += map.swapable ? '%' : '_';
		out += '	['+index+']'+(target ? '|>' : '  ')+'['+rel_dir+']'
			+(target ? '|>' : '  ')
			// +'['+gs_curr+(gs_maxed ? ' M' : ']')
			+' #'+entID+'  '
			+descrn+'	['+map.ve_colour+', '+map.lb_size+', '+map.ml_size+']'+(target ? '<==' : '')
//			  +' Ms: '+ map.staticMFD.toString(2)+' Md: '+map.dynamicMFD.toString(2)
			+' stat: '+ cd._number_str( map.staticMFD, 0, 2 ) + ' dyn: '+ cd._number_str( map.dynamicMFD, 0, 2)
			+ (map.have_scanned ? ' h_s: ' + map.have_scanned :'')
			+'\n';
	}
	out += 'ws.$SightingsMap.length = ' + mapping.length;
	log(out);
	if( thargoids ) log( 'Thargoids:\n' + thargoids );
}

this._rel2PS = function rel2PS( name, position, distance ) {
	var cd = worldScripts.telescope_debug;
	var ps = player && player.ship;
	var ps_position = ps.position;
	let posn =	position.constructor === Vector3D ? position :
				Array.isArray(position) ? new Vector3D( position ) :
				position.position || null;
	if( !posn ) {
		return 'missing position for ' + name;
	}
	let dist = distance || posn.distanceTo( ps_position );
	let padding = '________________';
	let pname = padding.slice( name.length - 16 ) + name;
	if( name === 'lightball' ) pname = '__' + pname;
	let vector = posn.subtract( ps_position );
	let fwd, right, up, rpt = '';
	fwd =	( ps.vectorForward.angleTo( vector ) *180/3.1415927 ).toFixed();
	right =	( ps.vectorRight.angleTo( vector ) *180/3.1415927 ).toFixed();
	up =	( ps.vectorUp.angleTo( vector ) *180/3.1415927 ).toFixed();
	rpt += pname + ' is ' + ( fwd <= 90 ? 'forward ' + fwd : 'astern ' + (fwd - 90) ) + '°, '
							+ ( right <= 90 ? '	 starboard ' + right : '		port ' + (right - 90)) + '°, '
							+ ( up <= 90 ? '	above '+ up : '	   below ' + (up - 90)) 
							+ '° at ' + cd._number_str( dist ) + ' m';
	// vector = vector.direction();
	// fwd =	ps.vectorForward.dot( vector ).toFixed(5);
	// right =	ps.vectorRight.dot( vector ).toFixed(5);
	// up =	ps.vectorUp.dot( vector ).toFixed(5);
	// rpt += '\n\t' + pname + ' is ' + ( fwd >= 0 ? 'forward ' : 'astern ' ) + ( Math.acos(fwd) *180/3.1415927 ).toFixed() + '°, '
							// + ( right >= 0 ? '  starboard ' : '		  port ') + ( Math.acos(right) *180/3.1415927 ).toFixed() + '°, '
							// + ( up >= 0 ? '	  above ' : '	 below ') +	 ( Math.acos(up) *180/3.1415927 ).toFixed()
							// + '° at ' + cd._number_str( dist ) + ' m';
	return rpt;
}

this._curr_S_report = function _curr_S_report() {
	var cd = worldScripts.telescope_debug;
	var cs = worldScripts.telescope.$curr_Sighting;
	var ps = player && player.ship;
	var scannerRange = ps.scannerRange;

	let map = cs.map;
	if( !map ) {
		log('telescope', '_curr_S_report, $curr_Sighting.map is ' + map );
		return;
	}
	let ent = cs.ent;
	let map_ent_dist = cs.map.ent_dist;
	let marker = cs.marker;
	let lightball = cs.map.lightball;
	let tmp = 'ps.speed = ' + ps.speed + ',	   viewDirection = ' + ps.viewDirection + ',  for ' + ent;
	tmp += '\n ' + cd._rel2PS( 'ent', ent, map.ent_dist );
	if( marker )
		tmp += '\n ' + cd._rel2PS( 'marker', marker	 );
	if( lightball )
		tmp += '\n ' + cd._rel2PS( 'lightball', lightball );
//	let ent_posn = map.rank === 'ukn'  ? map.last_posn : ent.position;

	let marker_dist = scannerRange - 499.6;
	if( map_ent_dist < scannerRange ) {
		if( map_ent_dist < marker_dist )
			marker_dist = map_ent_dist;
	}
//	let targ_dir = ent_posn.subtract( ps_position ).direction();
//	let targ_posn = ps_position.add( targ_dir.multiply( marker_dist ) );//.add( speed_adj );
//	tmp += '\n ' + cd._rel2PS( 'new marker', targ_posn, marker_dist );

	log('telescope', tmp );
}

// debugger ///////////////////////////////////////////////////////////////////////////////////////

/*	(function () {
WS.telescope.$cag = worldScripts.telescope_debug._debug_msgs();
WS.telescope.$cag.init();
})()	// */

this._debug_msgs = function _debug_msgs() {
	function break_line( str, limit ) {			// break str into lines, avoiding splitting on an '=' if poss.
		function trim( x ) {											// return index of 1st non-whitespace char.
			var space;
			var index = x;											// never modify arguments
			do {
				space = str[ index ];
				if( space !== ' ' && space !== '\t' ) break;
				index++;
			} while( index < str.length );
			return index;
		}
		function output( x ) {										// push indices for next line; return true if finished w/ str
			var index = x;											// never modify arguments
			strptr = trim( strptr );
			lines.push( [ strptr, index ] );
			index = trim( index );
			if( index + limit >= str.length ) {						// just tail end remains
				lines.push( [ index, -1 ] );
				return true;										// finished w/ str
			}
			strptr = index;		i = index + limit;		first = -1;		newline = -1;
			return false;											// still more to output
		}
		function _seek( dir ) {										// walk str in direction dir across whitespace; return presence of an '='
			var equals = false;
			do {													// scan in direction dir for end of whitespace or an '='
				if( i + dir < 0 || i + dir >= str.length ) break;	// protect limits
				i += dir;
				letter = str[ i ];
				if( letter === '=' ) equals = true;
			} while( letter === ' ' || letter === '\t' || letter === '=' );
			i -= dir;
			return equals;											// return presence of an '='
		}
		var first = -1;		var letter, i;		var newline = -1;
		var lines = [];		var strptr = 0;
		for( i = strptr + limit; i > strptr; i-- ) {
			if( newline === -1 ) {
				newline = str.indexOf( '\n', strptr );
				if( newline >= strptr && newline <= i )				// respect imbedded newlines
					if( output( newline + 1 ) ) break;
			} else
				newline = 0;										// only check once per output()
			letter = str[ i ];
			if( letter === ' ' || letter === '\t' ) {
				if( _seek( 1 ) || _seek( -1 ) ) {					// look fwd & back for '='s; found a bad break, ie. has '='
					if( first < 0 ) first = i;						// save the first, just in case
				} else {
					if( output( i ) ) break;						// i pts to 1st letter in break
					continue;
				}
			}
			if( i - strptr <= 20 ) {								// no good break bound; reasonable lower limit on line length
				if( first > -1 ) {									// use previously found (bad) break
					if( output( first ) ) break;
				} else {											// no break found, arbitrary cut using limit
					if( output( strptr + limit ) ) break;
				}
			}
		}
		return lines;
	}

	function insert( str, frame ) {				// insert str into frame's msg buffer
		var msg_len = frame.msg.length;
		var pad_len = frame.pad.length;
		var str_len = str.length;
		if( msg_len === 0 ) {										// pad the 1st line
			frame.len = pad_len;									// will be added on output but need to account for its length now
		}
		if( str_len + frame.len	 > DEBUG_CONSOLE_WIDTH ) {			// begin a new line
			var lines = break_line( str, DEBUG_CONSOLE_WIDTH - pad_len );
			var start, end, i;
			for( i = 0; i < lines.length; i ++ ) {
				[start, end] = lines[ i ];
				frame.msg += str.substring( start, ( end === -1 ? str_len : end ) ) + '\n';
			}
		} else {
			frame.msg += ( msg_len > 0 ? '	' : '' ) + str;
		}
		var nl_index = frame.msg.lastIndexOf( '\n', msg_len );
		if( nl_index > -1 ) {										// calc length of last line in msg
			frame.len = msg_len - nl_index;							// '\n' has length of 1
		} else {
			frame.len = msg_len;
		}
	}

	function is_watched_fn( dbgfn, str ) {		// check user supplied $fns_watched to determine if msg should proceed
		var caller = frame.caller;
		if( fns_watched.length === 0 ) return true;					// fn limiting not used
		if( frame === null ) {
			log( project_name, 'ERR: _debug, null frame: debug cmd outside start...stop sequence.  is_watched_fn:  '+dbgfn+':  str = '+str+'\n############');
			return true; // allow in hopes of locating errant cmd
		}
		if( !fns_watched.hasOwnProperty( caller ) ) return false;	// fn limiting is in play but fn missing -> deny by default
		if( fns_watched[ caller ] !== true )						// its in use, fn found but turned off
			return false;
		return true;												// fn in list & set true!
	}

	function trim_space( str ) {				// trim white space at start & end of line; -can't use .trim() as it removes line breaks!
		return str.replace(/^[ \t\u1680\u180e\u2000-\u200a\u202f\u205f\u3000\ufeff]+|[ \t\u1680\u180e\u2000-\u200a\u202f\u205f\u3000\ufeff]+$/g, '');
	}

	function change_breaks( line ) {			// collapse & convert all line breaks
		return line.replace(/[\n\r\f\v\u00a0\u2028\u2029]+/g, '<BR>');
	}

	function output() {							// called by stop to output or start to flush pre-existing msg
		if( DEBUG_OFF ) return;
		var incl_hd = frame.incl_hd;
		var msg = frame.msg;
		if( !incl_hd && !msg )										// unless directed to output hd, only log if there's a msg
			return;													// excl. header when no msg
		if( incl_hd && frame.out_len > 0 )							// a force_hd not necess, as frame already has output
			return;
		var head = frame.head;
		var out = curr_pad + '	' + frame.caller + ':: ';			// 1st line of output always starts w/ the caller (.len incr'd in .add)
		if( head ) out += head;										// .head contains a newline char
		out += msg;
		out = trim_space( out );
		out = change_breaks( out );
		out = out.replace( /(<BR>)+$/g, '' );						// trim any trailing line break
		out = out.replace( /(<BR>)+/g, '\n' + curr_pad + '	' );	// add padding to start of each line
		if( out.slice( - curr_pad.length ) === curr_pad )			// out ends w/ just a pad
			out = out.slice( 0, - curr_pad.length );
		if( frame.logging ) log( project_name, out );				// writes to console and log file
		else				log( out );								// writes only to console
		frame.out_len += out.length;
		frame.head = '';	msg = '';	frame.len = 0;
		frame.incl_hd = false;										// reset for next time
	}

	function add_msg( via, str, format ) {		// insert format str into msg buffer
		if( !is_watched_fn( via, str ) ) return;
		if( !str ) return;
		if( stack.length === 0 )
			log( project_name, 'ERR: debug, no debug frame: check for premature ".stop()" or add a ".start()" statement. .' + via + ':	str = ' + str +'\n############');
		insert( format, frame );
	}

	// add( str ) - append str to output msg
	function add( str ) {						// user fn to append str to msg
		if( DEBUG_OFF ) return;
		add_msg( 'add', str, str );
	}

	// begin( str ) - output str at start of a new line (prepends a newline)
	function begin( str ) {						// user fn to begin a new line w/ str
		if( DEBUG_OFF ) return;
		add_msg( 'begin', str, (frame.len > 0 ? '\n' + str : str) );
	}

	// end( str ) - output str at end of curr. line (appends a newline)
	function end( str ) {						// user fn to end the current line w/ str
		if( DEBUG_OFF ) return;
		add_msg( 'end', str, str +'\n' );
	}

	// log_file( str, [con_log], [caller] )	 - stand-alone user fn to add output to log file
	//										 - it ignores $DEBUG_OFF flag but terminates any further action if true
	//										 - con_log flag can be used to output to console as well, in curr. frame
	//										 - caller may be from any previous frame; if not found & con_log is true, output added to current frame
	function log_file( str, con_log, caller ) {	// output data to log file independent of the current frame
		log( project_name, caller + ', ' + str );					// writes to log file
		if( DEBUG_OFF ) return;										// debugger if off, so just log msg
		if( stack.length === 0 )	return;							// no active frame
		if( con_log === undefined ) return;
		if( caller === undefined ) {								// add str to curr. frame
			add( str );
			return;
		}
		var prev;
		var len = stack.length;
		for( var i = 0; i < len; i ++ ) {
			prev = stack[ i ].caller;
			if( prev === caller ) {
				insert( str, prev )									// add str to caller's frame
				return;
			}
		}
		add( str );													// failed to locate caller, add to curr. frame
	}

	function Frame( c ) {	this.caller = c;		this.head = '';			this.msg = '';
							this.pad = '';			this.out_len = 0;		this.len = 0;
							this.incl_hd = false;	this.force_hd = 1;		this.logging = false;		}
	var used_frames = [];
	var used_frame_num = 0;
	function free_frame( frame ) {
//log('telescope', 'free_frame, for caller = ' + frame.caller + ', pool totals ' + (used_frame_num + 1)+ ', on stack: ' + (stack.length - 1) );
		frame.caller = null;	frame.head = '';		frame.msg = '';
		frame.pad = '';			frame.out_len = 0;		frame.len = 0;
		frame.incl_hd = false;	frame.force_hd = 1;		frame.logging = false;
		used_frames.push( frame );
		used_frame_num++;
	}
	function alloc_frame( caller ) {
		var frame;
		if( used_frame_num > 0 ) {
			frame = used_frames.pop();
			used_frame_num--
			frame.caller = caller;
//log('telescope', 'alloc_frame for caller = ' + caller + ', allocating recycled frame, pool totals ' + used_frame_num + ', on stack: ' + (stack.length + 1) );
		} else {
			frame = new Frame( caller );
log('telescope', 'alloc_frame, for caller = ' + caller + ', allocating NEW frame, pool totals ' + used_frame_num + ', on stack: ' + (stack.length + 1) );
		}
		stack.push( frame );
		return frame;
	}

	// start( caller, [hd], [force_hd], [logging] )	 - open new debug level; caller is function's name (output toggled in $fns_watched)
	//												 - hd is str (heading) that is output only if more str's comes in
	//												 - force_hd, an int, over-rules this (default = 1; see Frame object).
	//												   this give some context when a normally silent fn outputs a msg,
	//												   by showing who called it. The higher the number, the farther up the frame stack
	//												   headers are forced to output
	//												 - logging is a boolean to also write to log file (vs. debug console only)
	function start( caller, hd,					// user fn to init a msg frame, start's hd string is not counted in len, so if
					force_hd, logging ) {		// no fn msg's between start & stop, no output
		if( DEBUG_OFF ) return;
		var len = stack.length;
		if( len > 0 ) {												// not the root frame
			output();												// flush any pending output
			var fn_name = caller !== undefined ? caller
											   : 'Anonomous' + stack.length;
			var prev;
			for( var i = len -1; i >= 0; i-- ) {					// check not already on stack (fix for GDERM bug)
				prev = stack[ i ].caller;
				if( prev === fn_name ) {
					free_frame( stack.splice( i, 1 )[0] );
					curr_pad = curr_pad.slice( 0, -PAD_LEN );		// just remove that one entry, as we know we're not recursing
				}
			}
		}
		frame = alloc_frame( caller );
		var tmp_pad;
		if( stack.length % 2 === 0 ) tmp_pad = '=========';			// alternate padding char's
		else						 tmp_pad = '+++++++++';
		frame.pad = tmp_pad.substr( 0, PAD_LEN );
		curr_pad += frame.pad;
		if( caller !== undefined ) {
			frame.caller = caller;
			frame.len += frame.caller.length;
		} else {
			log( project_name, 'ERR: _debug, invalid caller: function name needed for output to be formatted correctly. .start: str = ' + hd +'\n############');
		}
		if( frame === null ) {
			log( project_name, 'ERR: _debug, internal error: unable to create frame; stack len = ' + len + ', .start:  caller: '+caller+'(),  str = ' + hd +'\n############');
		}
		if( logging !== undefined )	 frame.logging = logging;
		if( force_hd !== undefined ) frame.force_hd = force_hd;
		if( hd !== undefined ) {
			tmp_pad = curr_pad;										// save current padding str and clear curr_pad
			curr_pad = '';											//	 so we don't put any padding on the header
			insert( hd + '\n', frame );								// a start header always ends its output line
			curr_pad = tmp_pad;										// restore current padding
		}
		frame.head = trim_space( frame.msg );						// move to head
		frame.msg = '';
		frame.len = 0;												// reset len as header doesn't count
	}

	// _stop( caller, [str], [nolog] ) - close curr. level or regress to it if events interruped sequence; nolog suppresses log file output,
	//									 over-ruling logging flag in start
	function stop( caller, str, nolog ) {		// user fn to terminate a frame, ouputting any pending messages
		if( DEBUG_OFF ) return;
var have_str = str && str.length > 0; // debug
		if( caller === undefined )
			log( project_name, 'ERR: _debug, missing caller: function name needed for output to be formatted correctly. .stop:	str = ' + str +'\n############');
		if( frame === null )
			log( project_name, 'ERR: _debug, no debug frame: add a "start()" statement. .stop:	caller: '+caller+'(), str = ' + str +'\n############');
		if( str && caller === frame.caller ) {
			insert( '\n' + str, frame );
have_str = false;
		}
		if( nolog !== undefined ) frame.logging = !nolog;			// nolog over-rules frame.logging
		if( stack.length > 0 ) {
			var msg_len = frame.msg.length;
			output();												// flush any pending output
			if( stack.length > 1 ) {								// not the root frame
				var force_hd = ( msg_len > 0 ? frame.force_hd : 0 );
				if( force_hd > 0 ) {								// set higher frame(s) to output
					var i = stack.length;
					while( i-- && force_hd > 0 ) {
						stack[ i ].incl_hd = true;
						force_hd--;
					}
				}
			}
			free_frame( frame );
			frame = stack.pop();
			if( str && caller === frame.caller ) {				// for when a .stop is skipped
				log( project_name, 'ERR: _debug, missed .stop(): ' + frame.caller + '() stopped while ' + caller + '() still active. .stop:	 str = ' + str +'\n############');
				insert( '\n' + str, frame );					// still output msg
have_str = false;
			}
			curr_pad = curr_pad.slice( 0, -PAD_LEN );
		} else {
			frame = alloc_frame( caller );							// re-init safely after having lost track!
		}
if( have_str ) log('telescope', 'debug.stop, failed to output: ' + str );
	}

	function reset() {							// reset msg stack; meant for debug console but is also a user fn
		var len;
		len = stack.length;
		while( len-- ) {
			free_frame( stack[ len ] );
		}
		stack.length = 0;						// this will clear all ref's to frame objects
		frame = null;
		PAD_LEN = 3;
		curr_pad = '';
	}
	function init( project ) {					// import local var's at runtime by calling this fn
		var wc = worldScripts.telescope_debug;
		if( project ) {
			project_name = wc.$project_name = project;
		} else {
			project_name = wc.$project_name;
		}
		fns_watched = wc.$fns_watched;
		DEBUG_CONSOLE_WIDTH = wc.$DEBUG_CONSOLE_WIDTH;
		DEBUG_OFF = wc.$DEBUG_OFF;
	}

	var frame = null;
	var stack = [];		// stack built of Frame's when debugging fns are nested
	var PAD_LEN = 3;	// max 9 unless you enlarge 2 padding str's in start
	var curr_pad = '';
	var project_name, fns_watched, DEBUG_CONSOLE_WIDTH, DEBUG_OFF;

	return {	 init:	init,
				  add:	add,
				begin:	begin,
				  end:	end,
			 log_file:	log_file,
				reset:	reset,
				start:	start,
				 stop:	stop
	};
}

// debugger ends //////////////////////////////////////////////////////////////////////////////////

this._reportError = function _reportError( err, func, parms, depth, goDeep ) {
	// constants - adjust as needed
	var FILE_LEN = 100;		// cut-off len for file spec.
	var FNAME_LEN = 40;		// cut-off len for function name
	var ARGS_LEN = 60;		// cut-off len for arguments string
	var STRING_LEN = 80;	// cut-off for long strings
	var IPAD = ' ';			// inside padding, eg. after array open bracket, before close bracket
/*	
	err		(required) value passed in the catch statement
	
	func	(required) function that caught the error
			- can also be a string with the function's name if the
			  function is not named
			  eg. this.startUp = function() { ...			-> pass 'startUp'
			      this.startUp = function startUp() { ...	-> pass startUp
			- using a named function has the advantage of reporting 
			  any function properties  
				
	parms	(optional) parameters passed to func
			- if there are multiple parms, pass them in an array
			- can also include function variables if you want more 
			  information dumped
			  eg. this.shipSpawned = function shipSpawned( ship ) { 
					var dist = player.ship.position.distanceTo( ship ); ...
			      -> pass as parms: [ship, dist] to have both displayed
				  
	depth	(optional) # levels to expose contents of parameters/properties
			- default is 1, use 0 to suppress overly long output
			  eg. if a parm is an array of 3 items, 0 will print "<array of 3>",
				  1 will print "[ item1, item2, item3 ]"
				  - if item3 is an object: "[ item1, item2, <object of ..> ]"
				  2 will print the contents of each item, in this case, 
				    expanding the object
				  3 will expand the object's properties, etc
				  (ditto for nested arrays)
			- NB: depth > 2 will likely produce large output, eg. ships
			- will accept a 2 item array if you want different depths for 
			  parameters vs. properties, ie. [ parm depth, prop depth ]
			  eg. [ 2, 1 ]
			
	goDeep	(optional) boolean indicating whether or not to test hasOwnProperty
			when exposing properties
			- default is false meaning inherited properties won't be shown
			- eg. if you dump player.ship, false shows 63 properties, while true 
				  shows 226! (player properties + ship properties + entity properties)
			- inherited properties are prefixed by a caret '^'
				
	example usage:
	=============
	
	this.relativeDirection = function relativeDirection( position, map ) {
		var that = relativeDirection;
		var ws = ( that.ws = that.ws || worldScripts.telescope );
		var ps = player && player.ship;
		
		try {
			var dist = map.ent.positon.distanceTo( ps );			<== position mis-spelled
			...
			return relative_dirn;
		} catch( err ) {
			log( 'telescope', ws._reportError( err, relativeDirection, [position, map], 1 ) );
			if( debug ) 
				throw err;									
		}
	}

	Latest.log output:
	=================
		
	18:48:27.293 [LogEvents]: Asteroid 28897 spawned at 413 km
	18:48:27.898 [telescope]:
	function relativeDirection() 	 caught: 	TypeError: map.ent.positon is undefined
		parameters: [ (70839.2, -96302.7, 664477), <object of 6> ]
		properties: { ws: <object of 169> }
		file: ../AddOns/Norby.cag.Telescope.oxp/Scripts/telescope.js
			line: 6204,	relativeDirection( [object Vector3D], [object Object] )
		file: ../AddOns/Norby.cag.Telescope.oxp/Scripts/telescope_debug.js
			line:   47,	_display_map( [object Object] )
	18:48:27.899 [script.javaScript.exception.unexpectedType]: ***** JavaScript exception (oolite-debug-console 1.89): TypeError: map.ent.positon is undefined
	18:48:27.899 [script.javaScript.exception.unexpectedType]:       ../AddOns/Basic-debug.oxp/Scripts/oolite-debug-console.js, line 1117.
*/
		
	function trim_str( str ) {
		var result, len = str.length;
		if( len === 0 ) 
			return '<empty string>';
		result = str.replace( /[\u180e\u2000-\u200a\u202f\u205f\u3000]+/g, ' ' );
		result = result.replace( /[\n]+/g, '\\n' ).replace( /[\t]+/g, '\\t' )
		result = '"' + (len > STRING_LEN ? result.substr(0, STRING_LEN) + ' ...' : result) + '"';
		return result
	}

	var padding = [];
	function mkSpacePad( count ) {
		if( typeof count === 'number' ) {
			padding.length = count + 1;
			return padding.join(' ');
		}
		return ' ';
	}
 
	function countObjKeys( obj, deep ) {	// Object.keys( obj ).length only counts hasOwnProperty ones
		var count = 0;						// deep overrides goDeep
		if( goDeep || deep ) {
			for( let prop in obj ) 
				if( prop )					// this is just to silence JSLint
					count++;
		} else {
			count = Object.keys( obj ).length;
		}
		return count;
	}
	
	function rptType( obj ) {
		if( Array.isArray( obj ) ) {
			let len = obj.length;
			return len > 0 ? '<array of ' + len + '>' : '[]';
		} else if( obj instanceof Script ) {
			return '[Script "' + obj.name + '" version ' + obj.version + ']';
		} else if( typeof obj === 'object' ) {
			let len = countObjKeys( obj, true );	// ignore goDeep when counting
			return len > 0 ? '<object of ' + len + '>' : '{}';
		} else {
			return obj;
		} 
	}
	
	function hasComplex( obj ) {
		for( let prop in obj ) {
			if( goDeep || obj.hasOwnProperty( prop ) ) {
				let item = obj[ prop ];
				if( Array.isArray( item ) || (typeof item === 'object' && item !== null) )
					return true;
			}
		}
		return false;
	}

	function showComplex( obj, recurse ) {		
		var isArray = Array.isArray( obj );
		var len = isArray ? obj.length : countObjKeys( obj );
		if( len === 0 ) return isArray ? '[]' : '{}';
		var index = 0, 
			str = (isArray ? '[' : '{') + IPAD, 
			strLen = str.length;
		var recursable = recurse > 0 && hasComplex( obj );
		for( let prop in obj ) {
			if( goDeep || obj.hasOwnProperty( prop ) ) {
				let item = obj[ prop ];
				let propStr = isArray ? '' : 
							(goDeep && !obj.hasOwnProperty( prop ) ? '^' : '') + prop + ': ';
				let propLen = propStr.length;
				str += propStr;
				if( recursable ) {
					if( index === 0 ) {
						outStarts.push( (outStarts.length > 0 
										? outStarts[outStarts.length-1] + propLen + strLen
										: strLen + propLen + strLen) );
					}
					str += fmt_parm( item, recurse );
					if( index < len - 1 ) {		// not the last one
						let inset = outStarts.length > 1 ? outStarts[outStarts.length-2] : strLen;
						str += ',\n' + mkSpacePad( indentLen + inset );
					} else {
						str += IPAD;
					}
				} else {
					str += hasComplex( item ) ? rptType( item ) : fmt_parm( item, 0 );
					str += index < len - 1 ? ', ' : IPAD;
				}
				index++;
			}
		}
		if( recursable && index ) outStarts.pop();
		return str + (isArray ? ']' : '}');
	}
	
	var outStarts = [];	// stack of running total of recursed insets
	var parents = [];	// check parm not in parents to avoid endless recursion
	function fmt_parm( parm, recurse ) {
		if( parents.indexOf( parm ) < 0 ) {
			parents.push( parm );
		} else  {
			return parm;
		} 
		var type = typeof parm;
		var str = '';
		if( parm === null ) {
			str += 'null';
		} else if( type === 'undefined' ) {
			str += 'undefined';
		} else if( type === 'string' ) {
			str += trim_str( parm );
		} else if( type === 'boolean' ) {
			str += (parm ? 'true' : 'false');
		} else if( type === 'function' ) {
			str += 'function ' + parm.name + '()';
		} else if( parm instanceof Script ) {
			str += '[Script "' + parm.name + '" version ' + parm.version + ']';
		} else if( parm instanceof Vector3D ) {
			str += 'Vector3D: (' + parm.x.toFixed() + ', ' 
					+ parm.y.toFixed() + ', ' + parm.z.toFixed() + ')';
		} else if( parm instanceof Quaternion ) {
			str += 'Quaternion: (' + parm.w.toFixed() + ' + ' + parm.x.toFixed() + 'i + ' 
					+ parm.y.toFixed() + 'j + ' + parm.z.toFixed() + 'k)';
		} else if( Array.isArray( parm ) ) {
			str += showComplex( parm, recurse <= 1 ? 0 : recurse - 1 );
		} else if( type === 'object' && parm ) {
			str += showComplex( parm, recurse <= 1 ? 0 : recurse - 1 );
		} else {
			str += rptType( parm );
		}
		parents.pop();
		return str;
	}

	var funcProps = {};
	function propsNotName( obj ) {
		if( typeof obj !== 'function' ) return 0;	// backwards compatibity
		for( let key in funcProps ) {				// reset object
			if( funcProps.hasOwnProperty( key ) )
				delete funcProps[ key ];
		}
		for( let key in obj ) {
			if( key !== 'name' )
				funcProps[ key ] = obj[ key ];
		}
		return Object.keys( funcProps ).length;
	}
	
	var parmsLabel = '\n    parameters: ';
	var indentLen = parmsLabel.length - 1;	// -1 for \n
	var fnName = typeof func === 'function' ? func.name : func; // backwards compatibity
	var rpt, parmMax, propMax,
		bonus = Array.isArray( parms ) ? 1 : 0;			// don't count parms being an array as recursion (+ 1)
	if( Array.isArray( depth ) ) {
		parmMax = (depth.length > 0 && typeof depth[ 0 ] === 'number' ? ~~(depth[ 0 ]) : 1) + bonus;
		propMax = (depth.length > 1 && typeof depth[ 1 ] === 'number' ? ~~(depth[ 1 ]) : 1) + bonus;
	} else {
		parmMax = propMax = (typeof depth === 'number' ? ~~(depth) : 1) + bonus;
	}
	if( err instanceof Error ) {
		rpt = '\nfunction ' + fnName + '() \t caught: \t' + err.name + ': ' + err.message;
	} else {		// for thrown strings (user defined errors)
		rpt = '\nfunction ' + fnName + '() \t caught: \t' + err;
	}
	if( parms ) {
		rpt += parmsLabel + fmt_parm( parms, parmMax );
	}
	if( propsNotName( func ) ) {
		parmsLabel = '\n    properties: ';
		indentLen = parmsLabel.length - 1;	// -1 for \n
		rpt += parmsLabel + fmt_parm( funcProps, propMax + 1 );	// + 1 as funcProps is an object
	}
	
	// err is the stack object with properties: message, fileName, lineNumber, stack, name
	//  - stack is a long string containing <function call>@<filename>:<line #> separated by
	//    '\n' for each call in the stack
	if( err && err.stack ) {
		var lastFile, parsed, frame, fnCall, args, file, line, pad;
		var stk = err.stack.split( /[\n\r]+/ ); // split on line breaks
		for( let i = 0, len = stk.length; i < len; i ++ ) {
			// stack line format: fn(parms)@../AddOns/.../script.js:123
			parsed = stk[ i ].match( /^\s*(\w+)\((.*?)\)@(.*?):(.*?)$/ );
			if( !parsed || parsed.length < 5 ) break;
			[frame, fnCall, args, file, line] = parsed;
			if( file && file !== lastFile ) {	// suppress repeat of same filename
				if( file.length > FILE_LEN ) 
					file = file.substring( file.length - FILE_LEN ) + '...';
				rpt += '\n    file: ' + file;
				lastFile = file;
			}
			pad = line < 10 ? '   ' : line < 100 ? '  ' : line < 1000 ? ' ' : '' ;
			rpt += '\n        line: ' + pad + line + ',	'; 
			if( fnCall.length > FNAME_LEN ) fnCall = fnCall.substring(0, FNAME_LEN) + '...';
			if( args.length > ARGS_LEN ) args = args.substring(0, ARGS_LEN) + '...';
			if( args.length ) 					// add spaces inside function's parenthices
				args = ' ' + args.replace( /,/g, ', ' ) + ' ';
			rpt += fnCall + '(' + args + ')';	
		}
	}
	return rpt;
}
	
this._number_str = function _number_str( n, fixed, base ) {
	var that = _number_str;
	var round = ( that.round = that.round || Math.round );
	var working = ( that.working = that.working || [] );

	var comma, index, wk, len, str;
	str = typeof n === 'string' ? parseFloat( n ) : n;
	if( !isFinite( str ) ) return n.toString();
	str = fixed > 0 ? str.toFixed( fixed ).toString( base || 10 )
					: round( str ).toString( base || 10 );
	len = str.length;
	working.length = wk = 0;
	for( index = 0; index < len; index++ ) working[ index ] = str[ index ];
	index = str.indexOf( '.' );
	comma = index >=0 ? (index > 0 ? index - 1 : 0) : len - 1;
	for( ; comma > 0 ; comma -= 3 ) {
		index = comma - 2;
		if( (n < 0 ? index-1 : index) <= 0 ) break;
		for( wk = len - 1; wk >= index; wk-- )
			working[ wk + 1 ] = working[ wk ];
		working[ index ] = ',';
		len = working.length;
	}
	return working.join( '' );
}

//	worldScripts.telescope_debug
//	worldScripts.telescope_debug._showProps( cs.map, 'map' )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', false,	false,	true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', true,		false,	true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', false,	true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', false,	true,	true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', true,		true,	true )
//	worldScripts.telescope_debug._showProps( cs.map, 'map', false,	2,		true )

//	worldScripts.telescope_debug._showProps =

this._showProps =	function _showProps( obj, objName, newLine, show_deep, expand_arrays, show_type ) {
	//										default:	true	true (1)		true			false
	function trim_str( str ) {
		var result, len = str.length;
		if( len === 0 ) 
			return '<empty string>';
		result = str.replace( /[\u180e\u2000-\u200a\u202f\u205f\u3000]+/g, ' ' );
		result = result.replace( /[ ]{3:}/g, '  ' );
		result = result.replace( /[\n]+/g, '\\n' ).replace( /[\t]+/g, '\\t' )
		// result = str.replace( /[\s]+/g, ' . ' );
		result = '"' + (len > 90 ? result.substr(0, 90) + ' ...' : result) + '"';
		return result
	}

	function mkPad( i, suppress ) {
		if( i <= 0 || suppress ) return '';
		padding.length = newLine ? i : 1;
		return (newLine ? '.' :'') + padding.join( '    ' ); // '.' needed for console, as leading space trimmed
	}
	 
	function rptType( obj, showIt ) {
		if( Array.isArray( obj ) )
			return show_type || showIt ? ' <array: ' + obj.length + ' elements> ' : ''; 
		else if( typeof obj === 'object' )
			return show_type || showIt ? ' <object: ' + Object.keys( obj ).length + ' keys> ' : ''; 
		else
			return ' <' + typeof obj + '> ';
	}
 
	function show_array( array, recurse ) {
		if( array.length === 0 ) return '[ ]';
		ilevel++;
		var str = '[ ';
		var pad = mkPad( ilevel, !expand_arrays );
		var len = array.length;
		for( var index = 0; index < len; index ++ ) {
			str += !expand_arrays || !newLine ? (index > 0 ? ', ' : '') : '\n';
			str += pad;
			if( expand_arrays )
				str += index + ': ' + fmt_prop( array[ index ], recurse );
			else
				str += fmt_prop( array[ index ], (recurse <= 1 ? 0 : recurse - 1) );
		}
		ilevel--;
		if( len > 0 && expand_arrays && newLine ) str += '\n';
		str += mkPad( ilevel, !len || !expand_arrays ) + ' ]' + rptType( array );
		return str;
	}

	function show_obj( obj, recurse ) {
		ilevel++;
		var str = '{ ';
		var pad = mkPad( ilevel );
		var len = Object.keys( obj ).length;
		for( var item in obj ) {
			str += newLine ? '\n' : '    ';
			str += pad;
			if( Array.isArray( obj ) ) {
				str += show_array( obj[ item ], recurse );
			} else {
				str += (obj.hasOwnProperty( item ) ? '' : '^') + item + ': ';
				str += fmt_prop( obj[ item ], recurse );
			}
			str += ';';
		}
		ilevel--;
		if( len > 0 && newLine ) str += '\n';
		str += mkPad( ilevel, !len ) + ' }' + rptType( obj );
		return str;
	}
	
	function fmt_prop( prop, recurse ) {
		if( parents.indexOf( prop ) < 0 )
			parents.push( prop );
		else 
			return prop;
		var type = typeof prop;
		var str = '';
		if( prop === null ) {
			str += 'null';
		} else if( type === 'undefined' ) {
			str += 'undefined';
		} else if( type === 'string' ) {
			str += trim_str( prop ) + (show_type ? ' <string: ' + prop.length + ' char.s>' : '');
		} else if( type === 'number' ) {
			str +=	cd._number_str( prop, prop % 1 === 0 ? 0 : expand_arrays ? 3 : 1 )
					+ (show_type ? ' <number>' : '');
		} else if( type === 'boolean' ) {
			str += prop + (show_type ? ' <boolean>' : '');
		} else if( type === 'function' ) {
			str += ' function ' + prop.name + '()';
		} else if( Array.isArray( prop ) ) {
			if( expand_arrays ) {
				str += show_array( prop, (recurse <= 1 ? 0 : recurse - 1) );
			} else {
				str += rptType( prop, true );
			}
		} else if( type === 'object' && prop ) {
			if( recurse > 0 ) {
				str += show_obj( prop, (recurse <= 1 ? 0 : recurse - 1) );
			} else {
				str += rptType( prop, true );
			}
		} else {
			str += prop + (show_type ? ' <'+type+'>' : '');
		}
		parents.pop();
		return str;
	}
	
	var that = _showProps;
	var cd = ( that.cd = that.cd || worldScripts.telescope_debug );
	
	var padding = [], parents = [];
	if( show_type === undefined ) 		show_type = false;
	if( expand_arrays === undefined )	expand_arrays = true;	
	if( show_deep === undefined ) 		show_deep = 1;
	if( newLine === undefined ) 		newLine = true;
	var rmax = !show_deep ? 0 : show_deep === true ? 1 : ~~(show_deep);
	var ilevel = 1;

	return (newLine ? '\n' : ' ') + objName + ': ' + fmt_prop( obj, rmax ) + (newLine ? '\n' : '');
}

}).call(this);














